use crate::parser::{Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::marker::PhantomData;
use std::rc::Rc;

pub struct MapParser<P, F, T1> {
    pub(crate) parser: P,
    pub(crate) mapper: F,
    pub(crate) phantom: PhantomData<T1>,
}

impl<TToken, T1, T2, TContext, P, F> Parser<TToken, T2, TContext> for MapParser<P, F, T1>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T1, TContext>,
    F: Fn(&T1) -> T2,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T2> {
        self.parser
            .parse(context, input)
            .map(|(remainder, result)| (remainder, (self.mapper)(&result)))
    }
}

pub struct BindParser<P, F, T1> {
    pub(crate) parser: P,
    pub(crate) binder: F,
    pub(crate) phantom_data: PhantomData<T1>,
}

impl<TToken, T1, T2, TContext, P, P2, F> Parser<TToken, T2, TContext> for BindParser<P, F, T1>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T1, TContext>,
    F: Fn(&T1) -> P2,
    P2: Parser<TToken, T2, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T2> {
        let (input, middle) = self.parser.parse(context.clone(), input)?;

        (self.binder)(&middle).parse(context, input)
    }
}

pub struct ConvertParser<P, T1, T2> {
    pub(crate) parser: P,
    pub(crate) phantom_data1: PhantomData<T1>,
    pub(crate) phantom_data2: PhantomData<T2>,
}

impl<TToken, T1, T2, TContext, P> Parser<TToken, T2, TContext> for ConvertParser<P, T1, T2>
where
    TToken: Debug + Clone,
    T2: From<T1>,
    P: Parser<TToken, T1, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T2> {
        self.parser
            .parse(context, input)
            .map(|(input, result)| (input, result.into()))
    }
}

pub struct NextParser<P, H, T> {
    pub(crate) parser: P,
    pub(crate) handler: H,
    pub(crate) phantom_data: PhantomData<T>,
}

impl<TToken, T, TResult, TContext, P, H> Parser<TToken, TResult, TContext> for NextParser<P, H, T>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    H: for<'f> Fn(ParserResult<'f, TToken, T>) -> ParserResult<'f, TToken, TResult>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, TResult> {
        let result = self.parser.parse(context, input);

        (self.handler)(result)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::parser::{parser_test_helper, take};
    use crate::text::char_parser;

    #[test]
    fn map_test() {
        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &(),
            take(5).map(|_| ()),
        )
    }

    #[test]
    fn bind_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'b',
            char_parser('a').bind(|_| char_parser('b')),
        );

        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'c',
            char_parser('a')
                .bind(|_| char_parser('b'))
                .bind(|_| char_parser('c')),
        );
    }
}
